package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.util.List;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DataSet;
import edu.stanford.nlp.ie.machinereading.structure.EntityMention;
import edu.stanford.nlp.ie.machinereading.structure.EventMention;
import edu.stanford.nlp.ie.machinereading.structure.ExtractionObject;
import edu.stanford.nlp.ie.machinereading.structure.Span;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.DocumentIdAnnotation;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.EntityMentionsAnnotation;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.EventMentionsAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.SentencesAnnotation;
import edu.stanford.nlp.pipeline.Annotation;
import edu.stanford.nlp.pipeline.StanfordCoreNLP;
import edu.stanford.nlp.stats.ClassicCounter;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.stats.Counters;
import edu.stanford.nlp.util.CoreMap;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;

/**
 * Implements a few tests to verify some annotation issues, i.e., overlapping entities and overlapping anchors
 */
public class AnnotationAnalysis {
  private static Logger logger = Logger.getLogger(TriggerClassifier.class.getName());
  private DataSet dataset;
  private StanfordCoreNLP processor = null;
  
  private AnnotationAnalysis(Properties props) {
    dataset = DataSet.load(props);
  }
  
  private void process(String dir) throws Exception {
    Annotation corpus = readCorpus(dir);
    analyzeEmbeddedEntities(corpus);
  }
  
  private void analyzeEmbeddedEntities(Annotation corpus) {
    // keeps track of embedded paths we've seen
    Counter<String> embeddedPaths = new ClassicCounter<String>();
    
    int totalEmbedded = 0, totalEmbeddedInEvents = 0;
    
    for(CoreMap sentence: corpus.get(SentencesAnnotation.class)) {
      String docid = sentence.get(DocumentIdAnnotation.class);
      List<EntityMention> entities = sentence.get(EntityMentionsAnnotation.class);
      List<EventMention> events = sentence.get(EventMentionsAnnotation.class);
      if(entities == null) continue;
      for(EntityMention e1: entities) {
        for(EntityMention e2: entities) {
          if(e2 == e1) continue;
          Span s1 = e1.getExtent();
          Span s2 = e2.getExtent();
          if(s1.start() <= s2.start() && s1.end() >= s2.end()){
            logger.severe("Found embedded entities in document " + docid + ": " + e2 + e1);
            embeddedPaths.incrementCount(e2.getType() + " in " + e1.getType());
            totalEmbedded ++;
            
            // does this embedded entity (e2) participate in any events? if so, it means we can't actually remove it
            if(events != null){
              for(EventMention event: events){
                for(ExtractionObject arg: event.getArgs()){
                  if(arg instanceof EntityMention && arg == e2){
                    totalEmbeddedInEvents ++;
                    logger.severe("Found embedded in event " + e2 + "\nEVENT:\n" + event);
                  }
                }
              }
            }
          }
        }
      }
    }
    
    List<Pair<String,Double>> paths = Counters.toDescendingMagnitudeSortedListWithCounts(embeddedPaths);
    for(Pair<String,Double> path: paths) logger.severe("Embedded path (" + path.second + ") " + path.first);
    logger.severe("Total embedded entities that serve as arguments in events: " + totalEmbeddedInEvents + "/" + totalEmbedded);
  }
  
  private Annotation readCorpus(String dir) throws Exception {
    // reader
    boolean useOfflineAnnotations = true;
    BioNLPFormatReader reader = new BioNLPFormatReader(useOfflineAnnotations, -1, true, false, false, dataset);
    reader.setProcessor(processor);

    // read the BioNLP format
    MachineReading.setConsoleLevel(Level.SEVERE);
    reader.setLoggerLevel(Level.SEVERE);
    Annotation corpus = reader.parse(dir);

    // logging
    logger.info("DONE reading the corpus");
    MachineReading.setConsoleLevel(Level.FINE);
    logger.setLevel(Level.FINE);

    return corpus;
  }
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    String trainDir = props.getProperty("train.dir");
    assert trainDir != null : "Training directory must be specified!";
    logger.setLevel(Level.FINE);
    
    AnnotationAnalysis analysis = new AnnotationAnalysis(props);
    analysis.process(trainDir);
    System.err.println("Done.");
  }
}
