package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.domains.bionlp.BioNLPFormatReader.FirstProbAnnotation;
import edu.stanford.nlp.ie.machinereading.domains.bionlp.BioNLPFormatReader.HasGoldAnnotation;
import edu.stanford.nlp.ie.machinereading.domains.bionlp.BioNLPFormatReader.ProbAnnotation;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DataSet;
import edu.stanford.nlp.ie.machinereading.msteventextractor.GENIA09DataSet;
import edu.stanford.nlp.ie.machinereading.structure.EntityMention;
import edu.stanford.nlp.ie.machinereading.structure.EventMention;
import edu.stanford.nlp.ie.machinereading.structure.ExtractionObject;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.DocumentIdAnnotation;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.EntityMentionsAnnotation;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.EventMentionsAnnotation;
import edu.stanford.nlp.io.FileSystem;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.CoreAnnotations.CharacterOffsetBeginAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.CharacterOffsetEndAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.SentencesAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TextAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TokensAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TreeAnnotation;
import edu.stanford.nlp.pipeline.Annotation;
import edu.stanford.nlp.process.Morphology;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.util.CoreMap;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;

/**
 * Converts the original BioNLP annotations to an extended form of the CoNLL format
 * This uses the original tokenization and syntactic annotation provided by the task organizers (through BioNLP09OfflineAnalysisReader)
 * The output of this code is fed into MSTParser by BioNLPDependencyReader
 * 
 * @author Mihai 
 * @author David 
 */
public class BioNLPToCoNLL {
  // fake root token and its label
  private static CoreLabel root = new CoreLabel();
  private static final String ROOT_LABEL = "root";
  
  private static Logger logger = Logger.getLogger(BioNLPToCoNLL.class.getName());
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    String input = props.getProperty("input");
    String output = props.getProperty("output");
    if(input == null || output == null){
      System.err.println("Use -input to specify the path to the dataset and -output to specify the output file name!");
      System.exit(1);
    }
    
    Level level = Level.SEVERE;
    MachineReading.setConsoleLevel(level);
    convert(new GENIA09DataSet(), props, input, -1, true, true, output, level);
  }
  
  public static void convert(DataSet dataset, Properties props, String inputDirectory, int beam, boolean keepGoldTriggers,
      boolean documentLevelParsing, String output, Level logLevel) throws Exception {
    convert(dataset, props, inputDirectory, beam, keepGoldTriggers, documentLevelParsing, output, logLevel, null);
  }
  
  public static void convert(DataSet dataset, Properties props, String inputDirectory, int beam, boolean keepGoldTriggers,
      boolean documentLevelParsing, String output, Level logLevel, String goldTriggerDirectory) throws Exception {
    // if the output file already exists do NOT overwrite
    File f = new File(output);
    if (FileSystem.existsAndNonEmpty(f)) {
      logger.severe("Refusing to overwrite existing file " + f.getAbsolutePath() + ". Will NOT re-generate CoNLL format for directory " + inputDirectory + ".");
      return;
    }
    
    // make sure we can create this file first so we can fail fast
    PrintStream os = new PrintStream(new FileOutputStream(output));
    
    // reader
    logger.severe("Converting corpus from " + inputDirectory);
    boolean useOfflineAnnotations = true;
    BioNLPFormatReader reader = new BioNLPFormatReader(useOfflineAnnotations, beam, keepGoldTriggers, documentLevelParsing, true, dataset);
    
    // handle some funky cases that show up when stacking
    boolean removeUnusedEntities = Boolean.valueOf(props.getProperty("remove.unused.entities", "true"));
    if (goldTriggerDirectory != null) {
      // TODO note that this currently assumes only one set of stacking data
      reader.setGoldTriggerDirectory(goldTriggerDirectory);
      removeUnusedEntities = false;
    }
    
    reader.setRemoveUnusedEntities(removeUnusedEntities);
    
    // logging
    reader.setLoggerLevel(logLevel);
    logger.setLevel(logLevel);
    
    // read the BioNLP format
    Annotation corpus = reader.parse(inputDirectory);
    boolean hasGold = corpus.get(HasGoldAnnotation.class);
    assert (corpus.get(SentencesAnnotation.class) != null) : "Didn't read any sentences in the corpus.";
    logger.severe("DONE reading the corpus");
    
    //
    // convert to CoNLL
    //
    
    root.setWord(ROOT_LABEL);
    
    // parse each individual sentence separately
    if (!documentLevelParsing) {
      for (CoreMap sent : corpus.get(SentencesAnnotation.class)) {
        StringBuilder buf = new StringBuilder();
        convertSentenceToCoNLL(buf, sent, reader, props, dataset);
        String deps = buf.toString();
        if (deps.length() > 0) {
          os.print("#SENT:");
          for (CoreLabel token : sent.get(TokensAnnotation.class)) {
            os.print(" " + token.word() + " " + token.tag() + " " + token.get(CharacterOffsetBeginAnnotation.class) + " " + token.get(CharacterOffsetEndAnnotation.class));
          }
          os.println();
          Tree tree = sent.get(TreeAnnotation.class);
          os.println("#TREE: " + tree.toString());
          os.println("#DOC: " + sent.get(DocumentIdAnnotation.class));
          os.print(buf);
          os.println();
        }
      }
    }
    // parse the entire document as a single sentence
    else {
      List<List<CoreMap>> docSents = breakIntoDocs(corpus.get(SentencesAnnotation.class));
      for(List<CoreMap> doc: docSents){
        StringBuilder buf = new StringBuilder();
        convertDocToCoNLL(buf, doc, reader, props, dataset);
        String deps = buf.toString();
        if(deps.length() > 0){
          for(CoreMap sent: doc){
            os.print("#SENT:");
            for(CoreLabel token: sent.get(TokensAnnotation.class)){
              os.print(" " + token.word() + " " + token.tag() + " " + token.get(CharacterOffsetBeginAnnotation.class) + " " + token.get(CharacterOffsetEndAnnotation.class));
            }
            os.println();
            Tree tree = sent.get(TreeAnnotation.class);
            os.println("#TREE: " + tree.toString());
          }
          os.println("#DOC: " + doc.get(0).get(DocumentIdAnnotation.class));
          os.print(buf);
          os.println();
        }
      }
    }
    
    os.close();
    
    String metaFile = output + ".meta";
    os = new PrintStream(new FileOutputStream(metaFile));
    os.println("hasGold " + hasGold);
    os.close();
    logger.severe("DONE!");
  }
  
  private static List<List<CoreMap>> breakIntoDocs(List<CoreMap> allSents) {
    List<List<CoreMap>> sentsByDocs = new ArrayList<List<CoreMap>>();
    List<CoreMap> crtDoc = null;
    String crtDocId = "NONE";
    for(CoreMap sent: allSents){
      String docid = sent.get(DocumentIdAnnotation.class);
      if(! docid.equals(crtDocId)){
        // save the crt doc
        if(crtDoc != null && crtDoc.size() > 0) sentsByDocs.add(crtDoc);
        // start a new one
        crtDoc = new ArrayList<CoreMap>();
        crtDocId = docid;
      }
      crtDoc.add(sent);
    }
    if(crtDoc != null && crtDoc.size() > 0) sentsByDocs.add(crtDoc);
    logger.severe("Found " + sentsByDocs.size() + " documents in this corpus.");
    return sentsByDocs;
  }
  
  private static void convertDocToCoNLL(StringBuilder outputString, List<CoreMap> sentences, BioNLPFormatReader reader, Properties props, DataSet dataset) {
    // this stores the dependencies to be written in CoNLL.  This map goes from
    //      entity -> (head of entity, dependency label) 
    // where dependency labels are event types for anchors and slot names for arguments
    Map<CoreLabel,Pair<CoreLabel,String>> dependencies = new IdentityHashMap<CoreLabel, Pair<CoreLabel,String>>();
    
    // this stores type information for entities:
    //      entity -> (index in orig sent, index in reduced sent, type)
    Map<CoreLabel,EntityInfo> entityTypes = new IdentityHashMap<CoreLabel, EntityInfo>();

    // this will fill in the dependencies and entityTypes tables for us
    convertEventMentions(sentences, dependencies, entityTypes, reader, props, dataset);
    logger.info("Found " + dependencies.size() + " dependencies.");
    for(CoreLabel mod: dependencies.keySet()){
      Pair<CoreLabel,String> dep = dependencies.get(mod);
      if(! dep.second.startsWith(ROOT_LABEL)) logger.info("Found dependency: " + mod.word() + "-" + dep.second + "->" + dep.first);
    }
    for(CoreLabel ent: entityTypes.keySet()){
      EntityInfo type = entityTypes.get(ent);
      logger.info("Entity: " + type.id + " " + ent.word() + " " + type.type + " " + type.sentence + " " + type.headIndexOrig + " " + type.headIndexReduced + " " + type.prob + " " + type.firstProb);
    }
    
    //
    // the reduced sentence will only have words which are anchors or arguments
    // convert all this to CoNLL format
    //
    for(CoreMap sentence: sentences){
      List<CoreLabel> tokens = sentence.get(TokensAnnotation.class);
      for (CoreLabel token : tokens) {
        Pair<CoreLabel,String> parentAndDep = dependencies.get(token);
        if (parentAndDep == null) {
          continue;
        }
        CoreLabel parent = parentAndDep.first();
        String dependencyName = parentAndDep.second();
        //outputString.append(parent.word() + " -" + dependencyName + "-> " + token.word() + "\n");

        boolean useLemmatizer = Boolean.valueOf(props.getProperty("use.lemmatizer", "false"));
        String word = token.word();
        String lemma = useLemmatizer ? Morphology.stemStatic(word, token.tag(), true).word() : word.toLowerCase();

        EntityInfo types = entityTypes.get(token);
        String coarseType = types.type;

        int parentIndex = -1;
        if (parent.equals(root)) {
          parentIndex = 0;
        } else {
          // position of the parent in the reduced sentence
          logger.info("Finding info for parent: " + parent);
          parentIndex = entityTypes.get(parent).headIndexReduced;
        }
        assert parentIndex >= 0;

        // ID FORM LEMMA CPOSTAG POSTAG FEATS (not used) HEAD DEPREL STARTTOKEN ENDTOKEN HEADTOKEN ENTITYID SENTENCE_POSITION TRIGGER_PROBS
        String[] pieces = { Integer.toString(types.headIndexReduced), 
            word,
            lemma, 
            coarseType, 
            coarseType, 
            "_",
            Integer.toString(parentIndex), 
            dependencyName, 
            "_", "_", // not used
            Integer.toString(types.startOrig), Integer.toString(types.endOrig), // start and end tokens in the original sentence,
            Integer.toString(types.headIndexOrig), 
            types.id,
            Integer.toString(types.sentence),
            types.prob + "," + types.firstProb // features contains prediction probability for triggers and probability of first choice (if beam used) 
        };

        String line = StringUtils.join(pieces, "\t") + "\n";
        outputString.append(line);
      }
    }
  }
  
  /**
   * Takes a sentence and writes to in CoNLL format to a StringBuilder.
   * Sensitive to the instance options like predictEntities.
   * 
   * @param outputString where to write the CoNLL formatted sentence
   * @param sentence the input sentence as a CoreMap. 
   */
  private static void convertSentenceToCoNLL(StringBuilder outputString, CoreMap sentence, BioNLPFormatReader reader, Properties props, DataSet dataset) {
    List<CoreLabel> tokens = sentence.get(TokensAnnotation.class);
    logger.info("Converting sentence: " + sentence.get(TextAnnotation.class));
    if (sentence.get(EntityMentionsAnnotation.class) != null) {
      logger.fine("Have the following entities in this sentence:");
      for (EntityMention e : sentence.get(EntityMentionsAnnotation.class)) {
        logger.fine(e.toString());
      }
    }
    if (sentence.get(EventMentionsAnnotation.class) != null) {
      logger.fine("Have the following events in this sentence:");
      for (EventMention e : sentence.get(EventMentionsAnnotation.class)) {
        logger.fine(e.toString());
        for (ExtractionObject p : e.getParents()) {
          logger.fine("\tParent is: " + p.toString());
        }
      }
    }
    
    // this stores the dependencies to be written in CoNLL.  This map goes from
    //      entity -> (head of entity, dependency label) 
    // where dependency labels are event types for anchors and slot names for arguments
    Map<CoreLabel,Pair<CoreLabel,String>> dependencies = new HashMap<CoreLabel, Pair<CoreLabel,String>>();
    
    // this stores type information for entities:
    //      entity -> (index in orig sent, index in reduced sent, type)
    Map<CoreLabel,EntityInfo> entityTypes = new HashMap<CoreLabel, EntityInfo>();

    // this will fill in the dependencies and entityTypes tables for us
    List<CoreMap> sentences = new ArrayList<CoreMap>();
    sentences.add(sentence);    
    convertEventMentions(sentences, dependencies, entityTypes, reader, props, dataset);
    for(CoreLabel mod: dependencies.keySet()){
      Pair<CoreLabel,String> dep = dependencies.get(mod);
      logger.info("Found dependency: " + mod.word() + "-" + dep.second + "->" + dep.first);
    }
    for(CoreLabel ent: entityTypes.keySet()){
      EntityInfo type = entityTypes.get(ent);
      logger.info("Entity: " + type.id + " " + ent.word() + " " + type.type + " " + type.headIndexOrig + " " + type.headIndexReduced + " " + type.prob + " " + type.firstProb);
    }

    // the reduced sentence will only have words which are anchors or arguments
    for (CoreLabel token : tokens) {
      Pair<CoreLabel,String> parentAndDep = dependencies.get(token);
      if (parentAndDep == null) {
        continue;
      }
      CoreLabel parent = parentAndDep.first();
      String dependencyName = parentAndDep.second();
      //outputString.append(parent.word() + " -" + dependencyName + "-> " + token.word() + "\n");

      EntityInfo entityInfo = entityTypes.get(token);
      String coarseType = entityInfo.type;

      int parentIndex = -1;
      if (parent.equals(root)) {
        parentIndex = 0;
      } else {
        // position of the parent in the reduced sentence
        logger.info("Finding info for parent: " + parent);
        parentIndex = entityTypes.get(parent).headIndexReduced;
      }
      assert parentIndex >= 0;
      
      assert entityInfo.startOrig <= entityInfo.headIndexOrig;
      assert entityInfo.headIndexOrig <= entityInfo.endOrig;
      
      // ID FORM LEMMA CPOSTAG POSTAG FEATS (not used) HEAD DEPREL STARTTOKEN ENDTOKEN HEADTOKEN ENTITYID SENTENCE_POSITION TRIGGER_PROBS
      String fineType = coarseType;

      boolean useLemmatizer = Boolean.valueOf(props.getProperty("use.lemmatizer", "false"));
      String word = token.word();
      String lemma = useLemmatizer ? Morphology.stemStatic(word, token.tag(), true).word() : word.toLowerCase();
      
      // sort of domain specific, but all of our domains include simple and complex categories
      if (Boolean.valueOf(props.getProperty("use.fine.types", "false"))) {
        if (dataset.getTypeGeneralizations().get("simple").contains(fineType)) {
          fineType = "simple";
        } else if (dataset.getTypeGeneralizations().get("complex").contains(fineType)) {
          fineType = "complex";
        }
      }
      String[] pieces = { Integer.toString(entityInfo.headIndexReduced), 
          word,
          lemma, 
          coarseType, 
          fineType, 
          "_",
          Integer.toString(parentIndex), 
          dependencyName, 
          "_", "_", // not used
          Integer.toString(entityInfo.startOrig), Integer.toString(entityInfo.endOrig), // start and end tokens in the original sentence,
          Integer.toString(entityInfo.headIndexOrig), 
          entityInfo.id,
          Integer.toString(entityInfo.sentence),
          entityInfo.prob + "," + entityInfo.firstProb // features contains prediction probability for triggers and probability of first choice (if beam used) 
      };

      String line = StringUtils.join(pieces, "\t") + "\n";
      outputString.append(line);
    }
  }
  
  private static void convertEventMentions(List<CoreMap> sentences,
      Map<CoreLabel, Pair<CoreLabel, String>> dependencies,
      Map<CoreLabel, EntityInfo> entityTypes,
      BioNLPFormatReader reader, Properties props, DataSet dataset) {
    
    logger.info(reader.annotationsSummaryAsString(sentences, "Annotations before conversion to CoNLL"));
    
    // first extract all entity and anchor mentions in this sentence
    Map<ExtractionObject, EntityInfo> ents = new IdentityHashMap<ExtractionObject, EntityInfo>();
    int sentencePosition = 0;
    for(CoreMap sentence: sentences){
      List<CoreLabel> tokens = sentence.get(TokensAnnotation.class);
      if (sentence.get(EventMentionsAnnotation.class) != null) {
        // extract anchors of known events
        for (EventMention em : sentence.get(EventMentionsAnnotation.class)) {
          if (!ents.containsKey(em.getAnchor())) {
            int headPos = BioNLPFormatReader.headPosition(em, tokens);
            EntityInfo ei = new EntityInfo();
            ei.isEvent = true;
            ei.isPredictedEvent = false;
            ei.sentence = sentencePosition;
            ei.headIndexOrig = headPos;
            ei.headIndexReduced = -1; // not set yet. will be set later
            // Use the anchor type as the type (i.e., POS in MST), not the event type!
            // This is important: anchor types are maintained consistently in the reader. Event types are not!
            ei.type = em.getAnchor().getType();
            ei.startOrig = em.getAnchor().getExtentTokenStart();
            ei.endOrig = em.getAnchor().getExtentTokenEnd();
            ei.id = em.getAnchor().getObjectId();
            ei.prob = em.getAnchor().attributeMap().containsKey(ProbAnnotation.class) ? em.getAnchor().attributeMap().get(ProbAnnotation.class) : 1;
            ei.firstProb = em.getAnchor().attributeMap().containsKey(FirstProbAnnotation.class) ? em.getAnchor().attributeMap().get(FirstProbAnnotation.class) : 0;
            
            try {
              assert (ei.startOrig <= headPos) : "headPos must be after startOrig";
              assert (headPos <= ei.endOrig) : "headPos must be before endOrig";
            } catch (AssertionError ae) {
              System.out.println("Sentence: " + sentence);
              System.out.println("Doc ID: " + sentence.get(DocumentIdAnnotation.class));
              System.out.println("EntityInfo: " + ei);
              throw ae;
            }

            ents.put(em.getAnchor(), ei);
            logger.finest("Default root dependency for event anchor: " + em.getAnchor());
          }
        }
      }
      if (sentence.get(EntityMentionsAnnotation.class) != null) {
        // extract other entities
        // note that this may include event anchors!
        //   either false positives from the trigger classifier, or triggers of events that were removed during the sanity check
        for (EntityMention em : sentence.get(EntityMentionsAnnotation.class)) {
          if (!ents.containsKey(em)) {
            int headPos = BioNLPFormatReader.headPosition(em, tokens);
            EntityInfo ei = new EntityInfo();
            ei.isEvent = dataset.getEventTypes().contains(em.getType()); 
            ei.isPredictedEvent = BioNLPFormatReader.isTrigger(em) && !BioNLPFormatReader.isGoldTrigger(em);
            ei.sentence = sentencePosition;
            ei.headIndexOrig = headPos;
            ei.headIndexReduced = -1; // not set yet. will be set later
            ei.type = em.getType();
            ei.startOrig = em.getExtentTokenStart();
            ei.endOrig = em.getExtentTokenEnd();
            ei.id = em.getObjectId();
            ei.prob = em.attributeMap().containsKey(ProbAnnotation.class) ? em.attributeMap().get(ProbAnnotation.class) : 1;
            ei.firstProb = em.attributeMap().containsKey(FirstProbAnnotation.class) ? em.attributeMap().get(FirstProbAnnotation.class) : 0;
            
            try {
              assert (ei.startOrig <= headPos) : "headPos must be after startOrig";
              assert (headPos <= ei.endOrig) : "headPos must be before endOrig";
            } catch (AssertionError ae) {
              System.out.println("Sentence: " + sentence);
              System.out.println("Doc ID: " + sentence.get(DocumentIdAnnotation.class));
              System.out.println("EntityInfo: " + ei);
              throw ae;
            }
            
            ents.put(em, ei);
            logger.finest("Default root dependency for entity: " + em);
          }
        }
      }
      sentencePosition ++;
    } // end loop over all sentences
    
    // order mentions in textual order
    Set<ExtractionObject> keys = ents.keySet();
    List<Pair<ExtractionObject, EntityInfo>> sortedEnts = new ArrayList<Pair<ExtractionObject, EntityInfo>>();
    for (ExtractionObject key : keys) {
      sortedEnts.add(new Pair<ExtractionObject, EntityInfo>(key, ents.get(key)));
    }
    Collections.sort(sortedEnts, new TextOrderingComparator());
    
    // set the positions in the reduced sentence (only ents)
    // also, populate entityTypes with the CoreLabels corresponding to the heads
    for (int i = 0; i < sortedEnts.size(); i++) {
      EntityInfo ei = sortedEnts.get(i).second;
      assert (ei.sentence >= 0);
      List<CoreLabel> tokens = sentences.get(ei.sentence).get(TokensAnnotation.class);
      ei.headIndexReduced = i + 1; // position in the reduced sentence
      CoreLabel h = tokens.get(ei.headIndexOrig);
      entityTypes.put(h, ei);
      logger.info("Entity " + h.word() + " " + ei);
    }
    
    // by default, all mentions have root as the head
    // we adjust mentions that participate in events later
    Set<CoreLabel> mentionHeads = entityTypes.keySet();
    for (CoreLabel mh : mentionHeads) {
      EntityInfo ei = entityTypes.get(mh);
      assert (ei != null);
      assert !dependencies.containsKey(mh);
      dependencies.put(mh, new Pair<CoreLabel, String>(root, ROOT_LABEL));
      logger.fine("Created root dependency for token: " + mh);
    }
    
    // we now adjust the dependencies for mentions that participate in events as args or anchors
    for (CoreMap sentence : sentences) {
      List<EventMention> eventMentions = sentence.get(EventMentionsAnnotation.class);
      if (eventMentions == null) {
        // no useful stuff in this sentence
        continue;
      }

      for (EventMention eventMention : eventMentions) {
        logger.info("eventMention: " + eventMention);
        logger.info("eventMention parent(s): " + eventMention.getParents());
        
        // by now we must have ONE parent for this event
        ExtractionObject parentObject = eventMention.getSingleParent(sentence);
        logger.info(eventMention.getObjectId() + " ^^^ " + (parentObject != null ? parentObject.getObjectId() : "ROOT"));

        // make mappings for anchor
        CoreLabel anchor = asCoreLabel(eventMention);
        assert(dependencies.containsKey(anchor));

        CoreLabel parent;
        if (parentObject != null) {
          parent = asCoreLabel(parentObject);
        } else {
          parent = root;
        }

        String depLabel = null;
        if (parentObject != null) {
          assert (parentObject instanceof EventMention);
        }
        EventMention parentEvent = (EventMention) parentObject;
        if (parentEvent != null) {
          for (int i = 0; i < parentEvent.getArgs().size(); i++) {
            ExtractionObject a = parentEvent.getArg(i);
            String an = parentEvent.getArgNames().get(i);
            if (a == eventMention) {
              depLabel = an;
              break;
            }
          }
          if (depLabel == null) {
            System.err.println("Cannot find link between arg " + eventMention.getObjectId() + " and parent " + parentEvent.getObjectId() + " in document " + eventMention.getSentence().get(DocumentIdAnnotation.class));
          }
          assert (depLabel != null);
        } else {
          depLabel = ROOT_LABEL;
          Pair<CoreLabel, String> dep = dependencies.get(anchor);
          if (dep != null) {
            depLabel = dep.second();
          }
        }
        logger.info("Creating anchor dependency: " + anchor + "-" + depLabel + "->" + parent);
        dependencies.put(anchor, new Pair<CoreLabel, String>(parent, depLabel));

        // make mappings for each argument
        logger.info("Converting to dependencies: " + eventMention);
        List<ExtractionObject> args = eventMention.getArgs();
        List<String> argNames = eventMention.getArgNames();
        for (int i = 0; i < args.size(); i++) {
          ExtractionObject arg = args.get(i);

          // if this is a subevent, we can skip it since we'll add it above
          if (arg instanceof EventMention) {
            continue;
          }

          logger.fine("arg " + i + ": " + arg.toString() + " " + argNames.get(i));
          String argName = argNames.get(i);

          CoreLabel argToken = asCoreLabel(arg);
          logger.fine("Found arg token: " + argToken);
          assert(dependencies.containsKey(argToken));
          logger.info("Creating arg dependency: " + argToken + "-" + argName + "->" + anchor);
          dependencies.put(argToken, new Pair<CoreLabel, String>(anchor, argName));
        }
      } // end loop over all sentence events
    } // end loop over all sentences
  }
  
  /*
   * Helper functions
   */
  private static CoreLabel asCoreLabel(ExtractionObject obj) {
    List<CoreLabel> tokens = obj.getSentence().get(TokensAnnotation.class);
    return tokens.get(BioNLPFormatReader.headPosition(obj, tokens));
  }
  
  /*
   * Inner classes
   */
  static class EntityInfo {
    /** entity id, if a known entity (i.e., given in the .a1 file) */
    String id;
    /** index of the entity head in the original sentence */
    int headIndexOrig;
    /** index in the reduced sentence */
    int headIndexReduced;
    /** entity type */
    String type;
    /** which is the original sentence for this entity */
    int sentence;
    /** span start in the original sentence */
    int startOrig;
    /** span end in the original sentence */
    int endOrig;
    /** prediction probability (for predicted triggers) */
    double prob;
    /** if beam used, this is the prediction probability for the first choice of the classifier */
    double firstProb;
    /** is this the anchor of an event? */
    boolean isEvent;
    /**
     * is this a predicted event? (i.e. trigger classifier predicted it was a
     * trigger but it's not actually a gold trigger
     */
    boolean isPredictedEvent;
    
    public String toString() {
      return "(" + id + ", " + headIndexOrig + ", " + headIndexReduced + ", " + type + ", " + sentence + ", " + startOrig + ", " + endOrig + ")";
    }
  };
  
  static class TextOrderingComparator implements Comparator<Pair<ExtractionObject, EntityInfo>> {
    @Override
    public int compare(
        Pair<ExtractionObject, EntityInfo> o1,
        Pair<ExtractionObject, EntityInfo> o2) {
      // in different sents
      if(o1.second.sentence < o2.second.sentence) return -1;
      if(o1.second.sentence > o2.second.sentence) return 1;
      // they are in the same sentence
      if(o1.second.headIndexOrig < o2.second.headIndexOrig) return -1;
      if(o1.second.headIndexOrig == o2.second.headIndexOrig) return 0;
      return 1;
    }
  }
}
