package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.io.File;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DependencyInstance;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DependencyReader;
import edu.stanford.nlp.ie.machinereading.msteventextractor.GENIA09DataSet;
import edu.stanford.nlp.util.StringUtils;

/**
 * Computes the amount of non-projectivity in our converted corpus
 * This uses the train.dir property
 */
public class CheckNonProjectivity {
  private static final String TRAINING_CONLL = "training.conll";
  private static Level LOG_LEVEL = Level.INFO;
  private static Level READER_LOG_LEVEL = Level.SEVERE;
  private static Logger logger = Logger.getLogger(BioNLPToCoNLL.class.getName());
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    System.err.println("Using the following properties:\n" + props);
    logger.setLevel(LOG_LEVEL);
    
    //
    // Sanity checks
    //
    // make sure the working directory exists
    File wdir = new File(props.getProperty("working.dir"));
    if(! wdir.exists() || ! wdir.isDirectory() || ! wdir.canWrite()){
      throw new RuntimeException("Working directory " + wdir.getAbsolutePath() + " must exist and be a writable directory!");
    }
    
    //
    // Step 1: convert the BioNLP directories to our extended CoNLL format
    //
    MachineReading.setConsoleLevel(READER_LOG_LEVEL);
    File f = new File(wdir.getAbsolutePath() + File.separator + TRAINING_CONLL);
    if(! f.exists()){
      // converting the training corpus
      int trainBeam = Integer.parseInt(props.getProperty("trigger.train.beam"));
      BioNLPToCoNLL.convert(new GENIA09DataSet(), props, 
          props.getProperty("train.dir"), 
          trainBeam,
          trainBeam >= 0 ? false : true,
              Boolean.parseBoolean(props.getProperty("document.level.parsing", "false")),
              wdir.getAbsolutePath() + File.separator + TRAINING_CONLL, READER_LOG_LEVEL);
      MachineReading.setConsoleLevel(LOG_LEVEL);
    }
    
    //
    // Step 2: read individual sentences and compute non-projectivity stats
    //
    DependencyReader reader = DependencyReader.createDependencyReader("BIONLP");
    reader.startReading(wdir.getAbsolutePath() + File.separator + TRAINING_CONLL);
    for(DependencyInstance sent; (sent = reader.getNext()) != null; ){
      sentCount ++;
      if(sentCount % 100 == 0) System.err.println("Processed " + sentCount + " sentences...");
      checkNonProjectivity(sent);
    }
    System.err.printf("Found %d out of %d non-projective dependencies.\n", nonProjDependencyCount, dependencyCount);
    System.err.printf("Found %d out of %d non-projective sentences.\n", nonProjSentCount, sentCount);
  }
  
  static int sentCount = 0;
  static int nonProjSentCount = 0;
  static int dependencyCount = 0;
  static int nonProjDependencyCount = 0;
  
  private static void checkNonProjectivity(DependencyInstance sent) {
    for(int mod = 0; mod < sent.heads.length; mod ++){
      int head = sent.heads[mod];
      System.err.println(mod + " -> " + head);
    }
    System.err.println();

    Set<Integer> alreadyCounted = new HashSet<Integer>(); 
    boolean foundNonProj = false;
    for(int mod = 0; mod < sent.heads.length; mod ++){
      if(alreadyCounted.contains(mod)) continue;
      int head = sent.heads[mod];
      dependencyCount ++;
      if(head == -1) continue;
      
      for(int innerMod = Math.min(mod, head) + 1; innerMod < Math.max(mod, head); innerMod ++){
        int innerHead = sent.heads[innerMod];
        if(innerHead < Math.min(mod, head) || innerHead > Math.max(mod, head)){
          nonProjDependencyCount ++;
          foundNonProj = true;
          System.err.println("Found non-projective dep: " + innerMod + " -> " + innerHead);
          alreadyCounted.add(innerMod);
        }
      }
    }
    if(foundNonProj){
      nonProjSentCount ++;
    }
  }
}
