package edu.stanford.nlp.ie.machinereading.domains.nfl;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.quiotix.html.parser.ParseException;

import edu.stanford.nlp.annotation.HtmlCleaner;
import edu.stanford.nlp.io.IOUtils;

/**
 * 
 * @author Andrey Gusev
 * @author David McClosky
 * 
 *         Fetches NFL game summary sentences from yahoo sports page
 * 
 */
public class NFLGameRecapCrawler {

  private static final int WAIT_TIME_BETWEEN_REQUESTS = 1000;
  private static final String DOMAIN_URL = "http://sports.yahoo.com";
	private static final String SCOREBOARD_URL = "/nfl/scoreboard?w=";

	private static final Pattern GAME_ID_PATTERN = Pattern.compile("/nfl/recap\\?gid=\\d+");
	// e.g. "<p>CANTON, Ohio (AP)--"
	private static final Pattern BYLINE_PATTERN = Pattern.compile("<p>(.*?)\\(AP\\)&#8212;");

	private static final String RECAP_URL_PREFIX = "/nfl/recap?gid=";
	private static final int START_WEEK = 1;
	private static final int END_WEEK = 22;
	private static final int PRESEASON_START_WEEK = 1;
	private static final int PRESEASON_END_WEEK = 5;

  private Logger logger = null;

  public static void main(String[] args) throws Exception {
    if (args == null || args.length < 1 || args.length > 2) {
      System.out.println("Usage: NFLGameRecapCrawler baseDirectoryForOutput [gameIdFile]");
      return;
    }

    File gameIdFile = null;
    if (args.length == 2) {
      gameIdFile = new File(args[1]);
    }
    
    NFLGameRecapCrawler crawler = new NFLGameRecapCrawler();
    crawler.crawlAndDumpToFiles(new File(args[0]), gameIdFile);
  }

	public NFLGameRecapCrawler() {
    logger = Logger.getLogger(NFLReader.class.getName());
	}

	public void crawlAndDumpToFiles(File baseDirectory, File gameIdFile) throws Exception {
	  Set<String> allGameIds;
	  if (gameIdFile == null) {
	    allGameIds = getAllGameIds(DOMAIN_URL + SCOREBOARD_URL);
	  } else {
	    allGameIds = new HashSet<String>();
	    for (String line : IOUtils.readLines(gameIdFile)) {
	      allGameIds.add(line);
	    }
	  }
		logger.info("Total games IDs: " + allGameIds.size());

		for (String gameId : allGameIds) {
		  logger.info("Game ID: " + gameId);
      String recapContent = sendGetRequest(DOMAIN_URL + RECAP_URL_PREFIX
          + gameId, "<div class=\"body_copy\">", "<li class=\"first email\">");
      
      recapContent = cleanupHtml(recapContent);
      logger.finest("recapContent: " + recapContent.trim());
      
      File outputFile = new File(baseDirectory, "nfl.game." + gameId + ".html");
      logger.info("Writing to " + outputFile.getAbsolutePath());
      FileWriter writer = new FileWriter(outputFile);
      writer.write(recapContent);
      writer.close();
   
      // don't hammer the server
      Thread.sleep(WAIT_TIME_BETWEEN_REQUESTS);
    }
	}

  private String cleanupHtml(String recapContent) throws IOException, ParseException, UnsupportedEncodingException {
    ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

    Set<String> acceptedTags = new HashSet<String>();
    acceptedTags.add("p");
    acceptedTags.add("a");

    HtmlCleaner htmlCleaner = new HtmlCleaner(new StringReader(recapContent), outputStream);
    htmlCleaner.setAcceptableTags(acceptedTags);
    htmlCleaner.setIgnoreComments(true);
    htmlCleaner.clean();

    String cleaned = outputStream.toString("UTF-8").trim();
    
    Matcher matcher = BYLINE_PATTERN.matcher(cleaned);
    if (matcher.find()) {
      String location = matcher.group(1).trim();
      cleaned = "<location>" + location + "</location>\n<p>" + matcher.replaceFirst(""); 
    }
    
    // does this stuff exist elsewhere?
    cleaned = cleaned.replaceAll("&#8216;", "`");
    cleaned = cleaned.replaceAll("&#39;", "'");
    cleaned = cleaned.replaceAll("&#8217;", "'");
    cleaned = cleaned.replaceAll("&#8220;", "``");
    cleaned = cleaned.replaceAll("&#8221;", "''");
    cleaned = cleaned.replaceAll("&#8212;", "--");
    cleaned = cleaned.replaceAll("&#8230;", "...");
    cleaned = cleaned.replaceAll("&hellip;", ""); // some sort of ellipsis that shows up in captiosn?
    
    cleaned = cleaned.replaceAll("AdChoices", "");
    cleaned = cleaned.replaceAll("\\(notes\\)", "");

    return cleaned;
  }

	/**
	 * Returns set of game ids based on week score pages
	 */
	private Set<String> getAllGameIds(String baseUrl) throws Exception {
		Set<String> gameIds = new HashSet<String>();
		
		// preseason games
    for (int week = PRESEASON_START_WEEK; week <= PRESEASON_END_WEEK; week++) {
      String content = sendGetRequest(baseUrl + "p" + week, null, null);
      gameIds.addAll(getGameIdsFromWeekScoreboard(content));
      // don't hammer the server
      Thread.sleep(WAIT_TIME_BETWEEN_REQUESTS);
      System.out.println("Got all ids for preseason week = " + week);
    }
		
    // normal season games
		for (int week = START_WEEK; week <= END_WEEK; week++) {
			String content = sendGetRequest(baseUrl + week, null, null);
			gameIds.addAll(getGameIdsFromWeekScoreboard(content));
			// don't hammer the server
			Thread.sleep(WAIT_TIME_BETWEEN_REQUESTS);
			System.out.println("Got all ids for normal season week = " + week);
		}

		return gameIds;
	}

	/**
	 * Extract game ids from from content of week score page
	 */
	private Set<String> getGameIdsFromWeekScoreboard(String content) {
		Matcher m = GAME_ID_PATTERN.matcher(content);

		Set<String> gameIds = new HashSet<String>();
		while (m.find()) {
			String match = m.group();
			gameIds.add(match.substring(RECAP_URL_PREFIX.length()));
		}

		return gameIds;
	}

	/**
	 * send get request and return page content optionally that appears between
	 * starting and ending lines
	 */
	private String sendGetRequest(String urlValue, String startingLine,
	    String endingLine) throws IOException {
		String result = null;
		BufferedReader rd = null;
		try {
		  logger.fine("Getting URL: " + urlValue);
			URL url = new URL(urlValue);
			URLConnection conn = url.openConnection();
			// Get the response
			rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
			StringBuilder sb = new StringBuilder();
			String line;
			boolean reachedStart = (startingLine == null);
			boolean reachedEnd = false;
			while ((line = rd.readLine()) != null) {
				if (reachedEnd) {
					break;
				}
				if (reachedStart) {
					sb.append(line);
					sb.append("\n");
				}

				reachedEnd = endingLine != null && line.contains(endingLine);
				reachedStart = reachedStart || line.contains(startingLine);
			}

			result = sb.toString();
		} finally {
			if (rd != null) {
				rd.close();
			}
		}
		return result;
	}
}
