package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.File;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import edu.stanford.nlp.process.Morphology;
import edu.stanford.nlp.util.ReflectionLoading;

/**
 * Contains all domain specific information/processing functions
 * @author David McClosky
 */

public abstract class DataSet {
  
  private Map<String, Set<String>> typeGeneralizations;
  private FilesystemInformation filesystemInformation;
  protected Set<String> entityTypes;
  protected Set<String> eventTypes;
  protected Set<String> zeroArgEvents;
  protected Set<String> singleThemeEvents;
  protected Set<String> eventsWithOnlyThemes;
  protected Set<String> eventsRequiringTheme;
  protected boolean splitBindingEvents;
  protected boolean discardEmbeddedEntities;
  
  public DataSet() {
    typeGeneralizations = buildTypeGeneralizations(); 
    zeroArgEvents = new HashSet<String>();
    discardEmbeddedEntities = false;
  }
  
  public boolean isZeroArgEvent(String type) {
    return zeroArgEvents.contains(type);
  }
 
  public Set<String> getEntityTypes() {
    return entityTypes;
  }
  
  public Set<String> getEventTypes() {
    return eventTypes;
  }
  
  public Map<String, Set<String>> getTypeGeneralizations() {
    return typeGeneralizations;
  }
  
  /*
   * Methods below can be implemented by subclasses and have default implementations.
   */
  
  /**
   * Simplify a relation name, e.g. "Theme" or "Cause2" (might want to remove the numbers after them, e.g.)
   */
  public String simplifyRelationName(String label) {
    if (label.toLowerCase().startsWith("theme")) {
      return "Theme";
    } else {
      return label;
    }
  }
  
  /**
   * Simplify an entity name, e.g. "Protein", "Regulation", "Gene_expression/Positive_regulation"
   * (joint entity names are allowed)
   */
  public String simplifyEntityType(String entityType) {
    return entityType;
  }
  
  /**
   * Return a set of labels which will be analyzed if error analysis is enabled in TriggerClassifier.
   */
  public Set<String> getLabelsForTriggerErrorAnalysis() {
    return new HashSet<String>();
  }
  
  /**
   * By default, this uses Morphology.stemStatic to stem words for use in TriggerClassification.
   */
  public String lemmatizeWordForTriggerClassification(String word, String pos) {
    return Morphology.stemStatic(word, pos).value();
  }
  
  /*
   * Methods below must be implemented by subclasses.
   */
  
  /**
   * Create an ontology for this dataset.
   * 
   * @return a map between generalized types and a set of its children.
   */
  public abstract Map<String, Set<String>> buildTypeGeneralizations();
  
  /**
   * Create word-word features between headIndex and childIndex for a given dependency instance.
   */
  public abstract void addDomainSpecificFeatures(DependencyPipe pipe, DependencyInstance instance, int headIndex, int childIndex, FeatureVector fv);
  
  /**
   * Create word-label features for a given node and connecting edge.
   */
  public abstract void addDomainSpecificLabeledFeatures(DependencyPipe pipe, DependencyInstance instance, int index, String labelName, boolean attachingRight, boolean childFeatures,
      FeatureVector fv);
  
  /**
   * Return true if a given argument role name should be included for this task/dataset combination.
   * @param role Argument role name (e.g. Theme, Cause, Site, AtLoc, etc.)
   */
  public abstract boolean includeRole(String role);

  /*
   * Methods (and a class) to help find dataset files on disk.
   */

  protected void setFilesystemInformation(FilesystemInformation filesystemInformation) {
    this.filesystemInformation = filesystemInformation;
  }

  public FilesystemInformation getFilesystemInformation() {
    return filesystemInformation;
  }

  public static abstract class FilesystemInformation {
    protected String syntacticParser = null;
    protected String tokenizer = null;
    protected String triggerDir = null;
    protected boolean useTriggerBeams = true;
    
    public FilesystemInformation(String tokenizer, String syntacticParser) {
      setTokenizer(tokenizer);
      setSyntacticParser(syntacticParser);
    }
    
    /*
     * Default implementations -- you should add error checking.
     */
    public void setTokenizer(String tokenizer) {
      this.tokenizer  = tokenizer;
    }
    public void setSyntacticParser(String syntacticParser) {
      this.syntacticParser = syntacticParser;
    }
    public void setTriggerDirectory(String triggerDir) {
      this.triggerDir = triggerDir; 
    }

    /*
     * Abstract methods
     */
    public abstract String getTokenizedFilename(String parentPath, String docId);
    public abstract String getTreeFilename(String parentPath, String docId);

    /**
     * Note that this default implementation ignores parentPath.  parentPath is just there
     * for compatibility with GENIA09.
     */
    public String getTriggerFilename(String parentPath, String docId, int beam) {
      String suffix;
      if (useTriggerBeams) {
        suffix = ".triggers.b" + beam;
      } else {
        suffix = ".a2";
      }
      String filename = docId + suffix;
      String[] triggerDirs = triggerDir.split(",");
      if (triggerDirs.length > 1) {
        // special logic for searching multiple directories
        for (String triggerDir : triggerDirs) {
          String fullFilename = triggerDir + File.separator + filename;
          if (new File(fullFilename).exists()) {
            return fullFilename;
          }
        }
      }
      
      // by default, just use the first one (if it exists or not)
      return triggerDirs[0] + File.separator + filename;
    }

    public void setUseTriggerBeams(boolean useTriggerBeams) {
      this.useTriggerBeams = useTriggerBeams;
    }
  }
  
  public boolean discardEmbeddedEntities() { return this.discardEmbeddedEntities; }

  public static DataSet load(Properties props) {
    String datasetClassName = props.getProperty("dataset.class");
    assert datasetClassName != null: "Please specify the dataset to use using the dataset.class property!";
    DataSet dataset = ReflectionLoading.loadByReflection(datasetClassName);
    
    String tokenizer = props.getProperty("dataset.tokenizer");
    if (tokenizer != null) {
      dataset.getFilesystemInformation().setTokenizer(tokenizer);
    }
    String parser = props.getProperty("dataset.parser");
    if (parser != null) {
      dataset.getFilesystemInformation().setSyntacticParser(parser);
    }
    String triggerDir = props.getProperty("dataset.trigger.dir");
    if (triggerDir != null) {
      dataset.getFilesystemInformation().setTriggerDirectory(triggerDir);
    }
    boolean useTriggerBeams = Boolean.valueOf(props.getProperty("dataset.use.trigger.beams", "true"));
    dataset.getFilesystemInformation().setUseTriggerBeams(useTriggerBeams);
    
    return dataset;
  }
}
