package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import edu.stanford.nlp.ling.StringLabelFactory;
import edu.stanford.nlp.trees.LabeledScoredTreeFactory;
import edu.stanford.nlp.trees.PennTreeReader;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.semgraph.SemanticGraph;
import edu.stanford.nlp.trees.semgraph.SemanticGraphFactory;

public class DependencyInstance implements Serializable {

  public static class SentenceInfo {
    // Tokens in the original sentence
    OrigToken[] originalSentence;
    
    // Parse tree of the original sentence as a string
    String originalTreeAsString;
    
    // Parse tree of the original sentence as a tree object
    Tree originalTree;
    
    // Dependencies in the original sentence
    SemanticGraph originalGraph;
    SemanticGraph originalGraphCollapsed;
  };

  private static final long serialVersionUID = 1L;
  public FeatureVector fv;
  public String actParseTree;

  // The various data types. Here's an example from Portuguese:
  //
  // 3 eles ele pron pron-pers M|3P|NOM 4 SUBJ _ _
  // ID FORM LEMMA COURSE-POS FINE-POS FEATURES HEAD DEPREL PHEAD PDEPREL
  //
  // We ignore PHEAD and PDEPREL for now.

  // FORM: the forms - usually words, like "thought"
  public String[] forms;

  // LEMMA: the lemmas, or stems, e.g. "think"
  public String[] lemmas;

  // COURSE-POS: the course part-of-speech tags, e.g."V"
  public String[] cpostags;

  // FINE-POS: the fine-grained part-of-speech tags, e.g."VBD"
  public String[] postags;

  // FEATURES: some features associated with the elements separated by "|", e.g.
  // "PAST|3P"
  public String[][] feats;

  // HEAD: the IDs of the heads for each element -- note that this will just be for the one-best parse
  public int[] heads;

  // DEPREL: the dependency relations, e.g. "SUBJ" -- note that this will just be for the one-best parse
  public String[] deprels;

  // K-best versions -- the outer array has dimension k so that together these items include the top k parses
  // parseScoresKBest includes the actual scores of each parse, nonProjectiveRearrangementsKBest reports how
  // many non-projective rearrangements were done in decoding (each of which will make parseScoresKBest less
  // accuracy, presumably)
  public int[][] headsKBest = null;
  public String[][] deprelsKBest = null;
  public double[] parseScoresKBest = null;
  public int[] nonProjectiveRearrangementsKBest = null;
  
  // RELATIONAL FEATURE: relational features that hold between items
  public RelationalFeature[] relFeats;
  
  // ID of the doc that contains this sentence
  public String docId;
 
  // info on the original sentence(s)
  protected List<SentenceInfo> sentences;
  
  // Start offsets in the original sentence
  protected int[] startInOriginal;
  // End offsets in the original sentence
  protected int[] endInOriginal;
  // Position of the head in the original sentence
  protected int[] headInOriginal;
  
  // Entity ids
  protected String[] ids;
  
  // Index of the sentence in the original space that contains this
  protected int[] sentenceIndex;
  
  // Predicted trigger probabilities (1.0 if gold trigger)
  protected double[] triggerProbs;
  // If this trigger was the second choice of the classifier, probability of first choice (0 otherwise)
  protected double[] firstTriggerProbs;
  
  public DependencyInstance(String[] forms, String[] postags, String[] labs, int[] heads,
      String docId,
      List<SentenceInfo> sentences, 
      int[] startOrigs,
      int[] endOrigs,
      int[] headOrigs,
      String[] ids,
      int[] sents,
      double[] tp,
      double[] ftp) {
    this.forms = forms;
    // TODO this isn't the Right Thing long term -- DependencyParser.parseInstance() calls this method but in doing so,
    // it loses lemma information which may be present.  It should stop doing that.
    this.lemmas = Arrays.copyOf(forms, forms.length);
    this.postags = postags;
    this.deprels = labs;
    this.heads = heads;
    
    this.docId = docId;
    this.sentences = sentences;
    
    this.startInOriginal = startOrigs;
    this.endInOriginal = endOrigs;
    this.headInOriginal = headOrigs;
    this.ids = ids;
    this.sentenceIndex = sents;
    assert (tp != null);
    this.triggerProbs = tp;
    assert (ftp != null);
    this.firstTriggerProbs = ftp;
  }
  
  public DependencyInstance(String[] forms, String[] lemmas, String[] cpostags,
      String[] postags, String[][] feats, String[] labs, int[] heads,
      RelationalFeature[] relFeats,
      String docId,
      List<SentenceInfo> sentences,
      int[] startOrigs,
      int[] endOrigs,
      int[] headOrigs,
      String[] ids,
      int[] sents,
      double[] tp,
      double[] ftp) {
    this.forms = forms;
    this.lemmas = lemmas;
    this.cpostags = cpostags;
    this.postags = postags;
    this.feats = feats;
    this.deprels = labs;
    this.heads = heads;
    this.relFeats = relFeats;
    
    this.docId = docId;
    this.sentences = sentences;
    convertOriginalTree();

    this.startInOriginal = startOrigs;
    this.endInOriginal = endOrigs;
    this.headInOriginal = headOrigs;
    this.ids = ids;
    this.sentenceIndex = sents;
    assert (tp != null);
    this.triggerProbs = tp;
    assert (ftp != null);
    this.firstTriggerProbs = ftp;
  }

  private void convertOriginalTree() {
    for (SentenceInfo sent : sentences) {
      try {
        sent.originalTree = new PennTreeReader(new StringReader(sent.originalTreeAsString), 
            new LabeledScoredTreeFactory(new StringLabelFactory())).readTree();
      } catch (IOException e) {
        // there's not really any I/O here, so it's possible this can't really happen
        e.printStackTrace();
        throw new RuntimeException("An IO error which really shouldn't have happened...happened.");
      }

      try {
        sent.originalGraph = SemanticGraphFactory.makeFromTree(sent.originalTree, false,
            false, true, false, true);
        sent.originalGraphCollapsed = SemanticGraphFactory.makeFromTree(sent.originalTree,
            true, true, true, false, true);
      } catch (Exception e) {
        String message = "Error converting tree: " + sent.originalTree.pennString();
        System.out.println(message);
        
        RuntimeException re = new RuntimeException(message);
        re.initCause(e);
        throw re;
      }
    }
  }
  
  public void setFeatureVector(FeatureVector fv) {
    this.fv = fv;
  }

  public int length() {
    return forms.length;
  }

  public String toString() {
    return toString(false);
  }
  
  public String toString(boolean extended) {
    StringBuffer sb = new StringBuffer();
    sb.append("docID: " + docId + "\n");
    sb.append("forms: " + Arrays.toString(forms)).append("\n");
    sb.append("ids: " + Arrays.toString(ids)).append("\n");
    sb.append("cpostags: " + Arrays.toString(cpostags)).append("\n");
    sb.append("heads: " + Arrays.toString(heads)).append("\n");
    sb.append("deprels: " + Arrays.toString(deprels)).append("\n");
    
    if (extended) {
      sb.append("startInOriginal: " + Arrays.toString(startInOriginal)).append("\n");
      sb.append("endInOriginal: " + Arrays.toString(endInOriginal)).append("\n");
      sb.append("headInOriginal: " + Arrays.toString(headInOriginal)).append("\n");
      sb.append("sentence: " + Arrays.toString(endInOriginal)).append("\n");
    }
    return sb.toString().trim();
  }

  private void writeObject(ObjectOutputStream out) throws IOException {
    out.writeObject(forms);
    out.writeObject(lemmas);
    out.writeObject(cpostags);
    out.writeObject(postags);
    out.writeObject(heads);
    out.writeObject(deprels);
    out.writeObject(actParseTree);
    out.writeObject(feats);
    out.writeObject(relFeats);
    out.writeObject(docId);
    out.writeObject(new Integer(sentences.size()));
    for (SentenceInfo sent : sentences) {
      out.writeObject(sent.originalSentence);
      out.writeObject(sent.originalTreeAsString);
      out.writeObject(sent.originalTree);
      out.writeObject(sent.originalGraph);
      out.writeObject(sent.originalGraphCollapsed);
    }
    out.writeObject(startInOriginal);
    out.writeObject(endInOriginal);
    out.writeObject(headInOriginal);
    out.writeObject(ids);
    out.writeObject(sentenceIndex);
    out.writeObject(triggerProbs);
    out.writeObject(firstTriggerProbs);
  }

  private void readObject(ObjectInputStream in) throws IOException,
      ClassNotFoundException {
    forms = (String[]) in.readObject();
    lemmas = (String[]) in.readObject();
    cpostags = (String[]) in.readObject();
    postags = (String[]) in.readObject();
    heads = (int[]) in.readObject();
    deprels = (String[]) in.readObject();
    actParseTree = (String) in.readObject();
    feats = (String[][]) in.readObject();
    relFeats = (RelationalFeature[]) in.readObject();
    docId = (String) in.readObject();
    sentences = new ArrayList<SentenceInfo>();
    Integer sentCount = (Integer) in.readObject();
    for (int i = 0; i < sentCount; i++) {
      OrigToken [] originalSentence = (OrigToken[]) in.readObject();
      String originalTreeAsString = (String) in.readObject();
      Tree originalTree = (Tree) in.readObject();
      SemanticGraph originalGraph = (SemanticGraph) in.readObject();
      SemanticGraph originalGraphCollapsed = (SemanticGraph) in.readObject();
      SentenceInfo si = new SentenceInfo();
      si.originalSentence = originalSentence;
      si.originalTreeAsString = originalTreeAsString;
      si.originalTree = originalTree;
      si.originalGraph = originalGraph;
      si.originalGraphCollapsed = originalGraphCollapsed;
      sentences.add(si);
    }
    startInOriginal = (int[]) in.readObject();
    endInOriginal = (int[]) in.readObject();
    headInOriginal = (int[]) in.readObject();
    ids = (String[]) in.readObject();
    sentenceIndex = (int[]) in.readObject();
    triggerProbs = (double[]) in.readObject();
    firstTriggerProbs = (double[]) in.readObject();
  }

}
