package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.File;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.stanford.nlp.ie.machinereading.domains.bionlp.BioNLPFormatReader;
import edu.stanford.nlp.process.Morphology;
import edu.stanford.nlp.util.StringUtils;

public class GENIA09DataSet extends DataSet {
  
  public class GENIA09FilesystemInformation extends FilesystemInformation {
    public GENIA09FilesystemInformation(String tokenizer, String syntacticParser) {
      super(tokenizer, syntacticParser);
    }
    
    public void setTokenizer(String tokenizer) {
      if (!(tokenizer.equals("stanford") || tokenizer.equals("default"))) {
        throw new IllegalArgumentException("No such tokenizer: " + tokenizer);
      }
      super.setTokenizer(tokenizer);
    }
    
    public String getTokenizedFilename(String parentPath, String docId) {
      String suffix = tokenizer.equals("stanford") ? ".stanford" : "";
      return parentPath + File.separator + docId + ".tokenized" + suffix;
    }
    
    public void setSyntacticParser(String parser) {
      if (parser.equals("mccc")) {
        parser = "stanford-mccc"; // normalize these two since they're the same for GENIA09
      }
      if (!(parser.equals("stanford-mccc") || parser.equals("default"))) {
        throw new IllegalArgumentException("No such parser: " + parser);
      }
      super.setSyntacticParser(parser);
    }
    
    public String getTreeFilename(String parentPath, String docId) {
      // we called our parser "stanford" (in terms of file extensions) but it's
      // really "mccc" (since in GENIA11, "stanford" is actually the stanford
      // parser and "mccc" is a different thing)
      String suffix = syntacticParser.equals("stanford-mccc") ? ".stanford" : "";
      return parentPath + File.separator + docId + ".pstree" + suffix;
    }

    public String getTriggerFilename(String parentPath, String docId, int beam) {
      String suffix;
      if (useTriggerBeams) {
        suffix = ".triggers.b" + beam;
      } else {
        suffix = "";
      }
      return parentPath + File.separator + docId + suffix;
    }
  }
  
  public GENIA09DataSet() {
    super();
    
    setFilesystemInformation(new GENIA09FilesystemInformation("stanford", "stanford-mccc"));
    
    entityTypes = new HashSet<String>(StringUtils.split("Protein"));
    eventTypes = new HashSet<String>(StringUtils.split("Gene_expression Transcription Protein_catabolism Localization Phosphorylation Binding Regulation Positive_regulation Negative_regulation"));
    singleThemeEvents = new HashSet<String>(eventTypes);
    singleThemeEvents.remove("Binding");
    eventsWithOnlyThemes = new HashSet<String>(StringUtils.split("Gene_expression Transcription Protein_catabolism Localization Phosphorylation Binding"));
    eventsRequiringTheme = new HashSet<String>(eventTypes);
    splitBindingEvents = true;
  }
  
  public String simplifyEntityType(String entityType) {
    // don't mess with entities
    if (!BioNLPFormatReader.isTrigger(entityType))
      return entityType;

    // keep these combinations. they are reasonably common
    if (entityType.equalsIgnoreCase("Gene_expression/Positive_regulation"))
      return entityType;
    if (entityType.equalsIgnoreCase("Gene_expression/Transcription"))
      return entityType;

    // break all other combinations and keep just the non-recursive type
    String[] bits = entityType.split("/");
    if (bits.length == 1) {
      return entityType; // not a combination
    } else if (bits.length == 2) {
      String b0 = bits[0].toLowerCase();
      String b1 = bits[1].toLowerCase();

      // if two regulations convert to one regulation
      if (b0.contains("regulation") && b1.contains("regulation")) {
        // keep the most common regulation: positive, then negative, then generic
        if (b0.contains("positive") || b1.contains("positive"))
          return "Positive_regulation";
        else if (b0.contains("negative") || b1.contains("negative"))
          return "Negative_regulation";
        return "Regulation";
      }

      // if one regulation, keep the other type
      if (b0.contains("regulation"))
        return bits[1];
      else if (b1.contains("regulation"))
        return bits[0];

      // if one protein, keep the other type
      if (b0.contains("protein"))
        return bits[1];
      else if (b1.contains("protein"))
        return bits[0];
    } 

    // more than 2 pieces (or we couldn't solve the two-piece combination) -- arbitrarily pick one
    return bits[0];
  }

  public boolean includeRole(String role) {
    String roleLower = role.toLowerCase();
    // just the Task 1 roles for now
    return (roleLower.startsWith("theme") || roleLower.startsWith("cause"));
  }
  
  public Map<String, Set<String>> buildTypeGeneralizations() {
    HashMap<String,Set<String>> typeGeneralizations = new HashMap<String, Set<String>>();
    Set<String> simpleEventTypes = new HashSet<String>(StringUtils.split("Gene_expression Transcription Protein_catabolism Localization Phosphorylation Binding"));
    Set<String> complexEventTypes = new HashSet<String>(StringUtils.split("Regulation Positive_regulation Negative_regulation"));
    typeGeneralizations.put("simple", simpleEventTypes);
    typeGeneralizations.put("complex", complexEventTypes);

    // simple + complex = non-protein    
    Set<String> nonProteinEventTypes = new HashSet<String>(simpleEventTypes);
    nonProteinEventTypes.addAll(complexEventTypes);
    typeGeneralizations.put("non-protein", nonProteinEventTypes);

    // degenerate generalization for proteins
    Set<String> proteinEventTypes = new HashSet<String>();
    proteinEventTypes.add("Protein");
    typeGeneralizations.put("protein", proteinEventTypes);
    return typeGeneralizations;
  }
  
  public void addDomainSpecificFeatures(DependencyPipe pipe, DependencyInstance instance, int headIndex, int childIndex, FeatureVector fv) {
    // non-regulation events must have proteins as arguments
    if (!isRegulation(instance, headIndex) && !isRoot(instance, headIndex)
        && !isProtein(instance, childIndex) && !isRoot(instance, childIndex)) {
      pipe.add("events:consistency:invalidargfornonreg", fv);
    }
  }
  
  public void addDomainSpecificLabeledFeatures(DependencyPipe pipe, DependencyInstance instance, int word, String relationName, boolean attachingRight, boolean childFeatures,
      FeatureVector fv) {
    // only regulation events can have cause edges
    if (!childFeatures && !isRoot(instance, word)
        && relationName.equalsIgnoreCase("cause")
        && !isRegulation(instance, word)) {
      pipe.add("events:consistency:invalidlabelfornonreg", fv);
    }
    
    // Proteins can not head an event
    if (!childFeatures && !isRoot(instance, word) && isProtein(instance, word)) {
      pipe.add("events:consistency:invalidevhead", fv);
    }
  }
  
  public Set<String> getLabelsForTriggerErrorAnalysis() {
    return new HashSet<String>(Arrays.asList("Regulation", "Negative_regulation", "Binding"));
  }
  
  private static final Pattern POSREG = Pattern.compile("up-?regulat", Pattern.CASE_INSENSITIVE);
  private static final Pattern NEGREG = Pattern.compile("down-?regulat", Pattern.CASE_INSENSITIVE);
  
  public String lemmaTizeWordForTriggerClassification(String word, String pos) {
    Matcher m = POSREG.matcher(word);
    if (m.find())
      return "positive-regulation";
    m = NEGREG.matcher(word);
    if (m.find())
      return "negative-regulation";

    return Morphology.stemStatic(word, pos).value();
  }
  
  /*
   * helper functions
   */
  private boolean isRegulation(DependencyInstance instance, int index) {
    if (instance.postags[index].toLowerCase().contains("regulation"))
      return true;
    return false;
  }

  private boolean isProtein(DependencyInstance instance, int index) {
    if (instance.postags[index].toLowerCase().contains("protein"))
      return true;
    return false;
  }

  private boolean isRoot(DependencyInstance instance, int index) {
    if (index <= 0)
      return true;
    return false;
  }
}
