package edu.stanford.nlp.util;

import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet;
import it.unimi.dsi.fastutil.objects.ReferenceSet;

import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * Utility functions for working with {@link CoreMap}'s.
 * 
 * @author dramage
 */
public class CoreMaps {
  
  /**
   * Returns a view of a collection of CoreMaps as a Map from each CoreMap to
   * the value it stores under valueKey. Changes to the map are propagated
   * directly to the coremaps in the collection and to the collection itself in
   * the case of removal operations.  Keys added or removed from the given
   * collection by anything other than the returned map will leave the map
   * in an undefined state.
   */
  public static <V, CM extends CoreMap, COLL extends Collection<CM>> Map<CM,V>
    asMap(final COLL coremaps, final Class<? extends TypesafeMap.Key<CoreMap, V>> valueKey) {
    
    final ReferenceSet<CM> references = new ReferenceOpenHashSet<CM>(coremaps);
    
    // an EntrySet view of the elements of coremaps
    final Set<Map.Entry<CM, V>> entrySet = new AbstractSet<Map.Entry<CM,V>>() {
      @Override
      public Iterator<Map.Entry<CM, V>> iterator() {
        return new Iterator<Map.Entry<CM,V>>() {
          Iterator<CM> it = coremaps.iterator();
          CM last = null;
          
          public boolean hasNext() {
            return it.hasNext();
          }

          public Map.Entry<CM, V> next() {
            final CM next = it.next();
            last = next;
            return new Map.Entry<CM,V>() {
              public CM getKey() {
                return next;
              }

              public V getValue() {
                return next.get(valueKey);
              }

              public V setValue(V value) {
                return next.set(valueKey, value);
              }
            };
          }

          public void remove() {
            references.remove(last);
            it.remove();
          }
        };
      }

      @Override
      public int size() {
        return coremaps.size();
      }
    };
    
    return new AbstractMap<CM,V>() {
      @Override
      public int size() {
        return coremaps.size();
      }
      
      @Override
      public boolean containsKey(Object key) {
        return coremaps.contains(key);
      }
      
      @Override
      public V get(Object key) {
        if (!references.contains(key)) {
          return null;
        }
        return ((CoreMap)key).get(valueKey);
      }
      
      @Override
      public V put(CM key, V value) {
        if (!references.contains(key)) {
          coremaps.add(key);
          references.add(key);
        }
        return key.set(valueKey, value);
      }
      
      @Override
      public V remove(Object key) {
        if (!references.contains(key)) {
          return null;
        }
        return coremaps.remove(key) ? ((CoreMap)key).get(valueKey) : null;
      }
      
      @Override
      public Set<Map.Entry<CM, V>> entrySet() {
        return entrySet;
      }
    };
  }
}
